# Loading the libraries
import numpy as np 
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import random
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.metrics import r2_score
plt.rcParams['figure.dpi'] = 300
plt.rcParams['savefig.dpi'] = 300
import warnings
warnings.filterwarnings('ignore')
from hbac_bias_detection.config import MODEL_MODE, print_mode


def pca_plot(data, title, alpha):
    """ Function to perform dimensionality reduction on the features, so that we can create 2-dimensional scatterplots.
    Takes as input the entire dataset, selects the features on which we want to cluster, and stores them in a temporary pd Dataframe. 
    This df is used to create a seaborn scatterplot. """
    
    pca_features = data.drop(['scaled_errors', 'predicted_value', 'true_value', 'errors', 'clusters', 'new_clusters'], axis=1)
    other_features = data[['scaled_errors', 'predicted_value', 'true_value', 'errors', 'clusters', 'new_clusters']]
    # pca_features = data.drop(['predicted_value', 'true_value', 'errors', 'clusters', 'new_clusters'], axis=1)
    # other_features = data[['predicted_value', 'true_value', 'errors', 'clusters', 'new_clusters']]

    df = pd.DataFrame(pca_features)
    # print(PCA(n_components=2).fit(df).explained_variance_ratio_) if print_mode else ''

    pca = pd.DataFrame(PCA(n_components=2).fit_transform(df), index=df.index)

    # pca = PCA(n_components=2, svd_solver='full')
    # pca.fit(X)
    # PCA(n_components=2, svd_solver='full')
    # print(pca.explained_variance_ratio_)

    # print(PCA(n_components=2).fit_transform(df))
    temp_dataset = pca.join(other_features, how='left')
    temp_dataset.rename( columns={0 :'Principal Component 1'}, inplace=True )
    temp_dataset.rename( columns={1 :'Principal Component 2'}, inplace=True )

    scatterplot = sns.scatterplot(data=temp_dataset, x='Principal Component 1', y='Principal Component 2', alpha=alpha, hue="clusters", size='errors', sizes=(1, 200), palette="tab10")
    scatterplot.set_title(title)
    scatterplot.legend(loc='center left', bbox_to_anchor=(1.0, 0.5), ncol=1)
    # plt.savefig('germancredit_kmeans.pngermancredit_kmeansg', dpi=300) # for saving a plot
    plt.show()

def initialize_dataset(raw_data, error_scaling_factor,with_errors=True, just_features=False, scale_features=True, with_classes=False):
    """ Initialisation of the dataset. Scales all the features and can also scale the errors, which can be included or excluded for clustering 
    It returns a scaled dataset with new columns "clusters" = 0 and "new_clusters" = -1, which are required for HBAC """

    new_data = raw_data.copy(deep=True)
    features = new_data.drop(['predicted_value', 'true_value', 'errors'], axis=1)
    if with_errors:
        scaling_factor = error_scaling_factor
        new_data['scaled_errors'] = StandardScaler().fit_transform(raw_data['errors'].values.reshape(-1, 1))
        new_data['scaled_errors'] = new_data['scaled_errors'] * scaling_factor

    if just_features:
        new_data = new_data.drop(['predicted_value', 'true_value', 'errors'], axis=1)

    if scale_features:
            to_scale = raw_data.drop(['predicted_value', 'true_value', 'errors'], axis=1).columns
            new_data[to_scale] = StandardScaler().fit_transform(features[to_scale])

    if with_classes:
        if MODEL_MODE == 'regression':
            to_scale = raw_data.columns
            new_data[to_scale] = StandardScaler().fit_transform(raw_data[to_scale])
        else:
            for col in ['predicted_value', 'true_value', 'errors']:
                new_data[col] = raw_data[col]

    new_data['clusters'] = 0
    new_data['new_clusters'] = -1
    return new_data

def get_random_cluster(clusters):
    ''' This function returns the value of a random cluster
    clusters Df.Column the column clusters '''
    result = -1
    while (result == -1):
        result = random.randint(0, len(clusters.unique()))
    print('This is the random cluster we have picked:', result) if print_mode else ''
    return result

def plot_clusters(data):
    ''' This function returns a scatterplot of the clusters ''' 
    scatterplot = sns.scatterplot(data=data, x='1st', y='2nd', hue="clusters", size='errors', sizes=(100, 20), palette="tab10")
    plt.show()

def accuracy(results):
    ''' This function calculates the accuracy of a DF dataframe
    It requires a df.column named "errors" '''
    if len(results) == 0:
        print("You are calculating the accuracy on a empty cluster") if print_mode else ''
        return 0    #to prevent bug if 1 cluster
    if MODEL_MODE == 'regression':
        # Using R-squared as acc
        # if len(results)==0:
        #     acc = 0
        # else:
        acc = r2_score(results['true_value'], results['predicted_value'])
        #acc = r2_score(results['true_value'], results['predicted_value'])
        # try:
        #     acc = r2_score(results['true_value'], results['predicted_value'])
        # except:
        #     print('R2 over empty set')
        #     acc = 1
    if MODEL_MODE == 'classification':
        correct = results.loc[results['errors'] == 0]
        acc = len(correct)/len(results)
    return acc

def bias_acc(data, cluster_id, cluster_col):
    ''' This function calculates the negative bias, which is the accuracy of the selected cluster - the accuracy of the remaining clusters 
    Cluster col: the name of the DF column where the cluster assignments are '''
    cluster_x = data.loc[data[cluster_col] == cluster_id]
    if len(cluster_x) ==0:
        print("This is an empty cluster! cluster ", cluster_id) if print_mode else ''
    remaining_clusters = data.loc[data[cluster_col] != cluster_id]
    if len(remaining_clusters) == 0:
        print("This cluster is the entire dataset. Cluster: ", cluster_id) if print_mode else ''
    return accuracy(cluster_x) - accuracy(remaining_clusters)

def bias_acc_towards_global_average(data, cluster_id, cluster_col, ave_acc):
    ''' This function calculates the negative bias, which is the accuracy of the selected cluster - the average accuracy
    Cluster col: the name of the DF column where the cluster assignments are '''
    cluster_x = data.loc[data[cluster_col] == cluster_id]
    return accuracy(cluster_x) - ave_acc

def get_next_cluster(data):
    ''' This function returns the cluster number with the highest variance. The variance is calculated based on the errors of each cluster.
    The cluster with the highest variance will be selected as splitting cluster
    The function requires df.columns named "clusters" and "errors '''
    n_cluster = max(data['clusters'])
    highest_variance = -1
    cluster_number = 0

    for i in data['clusters'].unique():
        if (i == -1):
            continue
        cluster_i = data.loc[data['clusters'] == i]
        variance_cluster = np.var(cluster_i['errors'])
        
        if variance_cluster > highest_variance:
            highest_variance = variance_cluster
            cluster_number = i
            print('--> Cluster with the highest variance:', cluster_number) if print_mode else ''

    return cluster_number

def get_max_abs_bias(fulldata, function=bias_acc):
    ''' This function returns the absolute value of highest negative bias of the newly introduced clusters 
    fulldata (DataFrame) should include a column new_clusters '''
    max_abs_bias = -999999
    for cluster_number in fulldata['new_clusters'].unique():
        if cluster_number == -1: #Outliers in DBScan
            continue
        current_bias = abs(function(fulldata, cluster_number, "new_clusters")) # abs function
        if current_bias > max_abs_bias:
            max_abs_bias = current_bias
    return max_abs_bias

def get_max_negative_bias(fulldata, function=bias_acc):
    ''' This function returns the highest negative bias of the newly introduced clusters 
    fulldata (DataFrame) should include a column new_clusters  --> used for identifying underperformed clusters '''
    max_neg_bias = 999999
    # print('potential new clusters: ', fulldata['new_clusters'].unique()) if print_mode else ''
    for cluster_number in fulldata['new_clusters'].unique():
        if cluster_number == -1: #Outliers in DBScan
            continue
        # current_bias = (function(fulldata, cluster_number, "new_clusters")) # abs function
        current_bias = (function(fulldata, cluster_number, "new_clusters")) # abs function
        if current_bias < max_neg_bias:
            max_neg_bias = current_bias
    print('max discriminating bias is:', max_neg_bias) if print_mode else ''
    return max_neg_bias

def get_max_bias_cluster(fulldata, printing=False, function=bias_acc):
    ''' This function returns the cluster linked to the highest negative bias of the newly introduced clusters 
    fulldata (DataFrame) should include a column new_clusters '''
    # max_abs_bias = 100 changed to max_abs_bias
    max_neg_bias = 100
    best_cluster = -2
    for cluster_number in fulldata['clusters'].unique():
        if cluster_number == -1: # Outliers in DBScan, which are excluded
            continue
        current_bias = (function(fulldata, cluster_number, "clusters")) # abs function to find the highest bias
        print(f"{cluster_number} has bias {current_bias}") if printing else ''
        if current_bias < max_neg_bias:
            max_neg_bias = current_bias
            best_cluster = cluster_number
    print('cluster with the highest discriminating bias:', best_cluster) if printing else ''
    return best_cluster, max_neg_bias

def calculate_variance(data):
    ''' This function calculates the variance for a DF. It requires a df.column named "clusters" and it uses the bias_acc_towards_global_avg '''
    variance_list_local = []
    for j in data['clusters'].unique():
        average_accuracy = accuracy(data)
        neg_bias_clus = bias_acc(data, j, 'clusters') 
        variance_list_local.append(neg_bias_clus) #variance_list was "neg_bias_list" before
    variance = np.var(variance_list_local) #this was "neg_bias_list"
    return variance

def get_min_cluster_size(data):
    ''' Returns the size of the smallest new cluster '''
    min_cluster_size = len(data)
    for i in data['new_clusters'].unique():
        # exclude the cluster -1 from being seen as a cluster, since it contains outliers
        if i == -1:
            continue
        size = len(data.loc[data['new_clusters']==i])
        if size < min_cluster_size:
            min_cluster_size = size
    return min_cluster_size



##### HBAC KMeans Algo
# def hbac_kmeans(data, max_iter=30, show_plot=True):
#     clus_model_kwargs = {
#         "n_clusters": 2,
#         "init": "k-means++",
#         "n_init": 10,
#         "max_iter": 300,
#     }
#
#     full_data = initialize_dataset(data)
#
#
#     x = 0 # initial cluster number
#     initial_bias = 0
#     if show_plot:
#         pca_plot(full_data, 'HBAC-KMeans', 0.6)
#     variance_list = []
#     average_accuracy = accuracy(full_data)
#     minimal_splittable_cluster_size = round(0.05 * len(full_data))
#     minimal_acceptable_cluster_size = round(0.03 * len(full_data))
#     print("average_accuracy is: ", average_accuracy)
#
#     for i in range(1, max_iter):
#         if i != 1:
#             variance_list.append(calculate_variance(full_data))
#         full_data['new_clusters'] = -1
#         candidate_cluster = full_data.loc[full_data['clusters'] == x]
#
#         if len(candidate_cluster) < minimal_splittable_cluster_size:
#             x = get_random_cluster(full_data['clusters'])
#             continue
#
#         # Apply KMeans
#         kmeans_algo = KMeans(**clus_model_kwargs).fit(candidate_cluster.drop(['clusters', 'new_clusters', 'predicted_value', 'true_value', 'errors'], axis=1))
#
#         candidate_cluster['new_clusters'] = pd.DataFrame(kmeans_algo.predict(candidate_cluster.drop(['clusters', 'new_clusters', 'predicted_value', 'true_value', 'errors'], axis=1)),index=candidate_cluster.index)
#         full_data['new_clusters'] = candidate_cluster['new_clusters'].combine_first(full_data['new_clusters'])
#
#         max_discr_bias = get_max_negative_bias(full_data) # was get_max_abs_bias, but now it only finds the discriminated clusters
#         min_new_size = get_min_cluster_size(full_data)
#
#         if (max_discr_bias <= initial_bias) & (min_new_size > minimal_acceptable_cluster_size): #abs: >
#             # Add new cluster
#             print('adding a new cluster')
#             n_cluster = max(full_data['clusters'])
#             full_data['clusters'][full_data['new_clusters'] == 1] =  n_cluster + 1
#
#             if show_plot:
#                 pca_plot(full_data, 'HBAC-KMeans', 0.6)
#             x = get_next_cluster(full_data)
#             initial_bias = max_discr_bias
#         else:
#             x = get_random_cluster(full_data['clusters'])
#
#     print('MAX_ITER')
#     print(variance_list)
#     return full_data